from enum import Enum

from ..lib.auto_repl import auto

class TokenTypes(Enum):
    ALL = auto()
    AND = auto()
    ASSIGN = auto()
    AT = auto()
    BEGIN = auto()
    BREAKPOINT = auto()
    CYCLE = auto()
    DEFINE = auto()
    ELSE = auto()
    END = auto()
    EOF = auto()
    EXPRESSION = auto()
    FROM = auto()
    GET = auto()
    GROUP = auto()
    IF = auto()
    LITERAL_STRING = auto()
    LOCATION = auto()
    LOGICAL = auto()
    NAME = auto()
    NUMBER = auto()
    OFF = auto()
    ON = auto()
    OR = auto()
    POWER = auto()
    PAUSE = auto()
    RANGE = auto()
    RAW = auto()
    REGISTER = auto()
    REPEAT = auto()
    SET = auto()
    SYNTAX_ERROR = auto()
    TIME_PATTERN = auto()
    TO = auto()
    UNITS = auto()
    UNTIL = auto()
    UNKNOWN = auto()
    WHILE = auto()
    WITH = auto()
    WAIT = auto()
    ZONE = auto()

    @classmethod
    def commands(cls):
        return (TokenTypes.ASSIGN,
                TokenTypes.GET,
                TokenTypes.OFF,
                TokenTypes.ON,
                TokenTypes.POWER,
                TokenTypes.PAUSE,
                TokenTypes.REGISTER,
                TokenTypes.SET,
                TokenTypes.UNITS,
                TokenTypes.WAIT)



#!/usr/bin/env python

import argparse
import logging

from bardolph.lib import injection
from bardolph.lib import settings
from bardolph.controller import arg_helper
from bardolph.controller import config_values
from bardolph.controller import light_module
from bardolph.controller.units import UnitMode
from bardolph.vm import machine
from bardolph.vm.instruction import Instruction, OpCode
from bardolph.vm.vm_codes import JumpCondition, LoopVar, Operand, Operator
from bardolph.vm.vm_codes import Register, SetOp

assembly = [
    #instructions

]

params_0 = (OpCode.BREAKPOINT, OpCode.COLOR, OpCode.END_LOOP, OpCode.LOOP,
            OpCode.NOP, OpCode.PAUSE, OpCode.POWER, OpCode.STOP, OpCode.WAIT)

params_1 = (OpCode.END, OpCode.GET_COLOR, OpCode.JSR, OpCode.OP, OpCode.POP,
            OpCode.PUSH, OpCode.PUSHQ, OpCode.ROUTINE)

def get_assembly():
    current_instruction = 0
    while current_instruction < len(assembly):
        value = assembly[current_instruction]
        current_instruction += 1
        yield value

def build_instructions():
    program = []
    it = iter(get_assembly())
    op_code = next(it, None)
    while op_code is not None:
        if op_code in params_0:
            param_count = 0
        elif op_code in params_1:
            param_count = 1
        else:
            param_count = 2
        param0 = None if param_count < 1 else next(it)
        param1 = None if param_count < 2 else next(it)
        program.append(Instruction(op_code, param0, param1))
        op_code = next(it, None)
    return program

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '-d', '--debug', help='do debug-level logging', action='store_true')
    ap.add_argument(
        '-f', '--fakes', help='use fake lights', action='store_true')
    arg_helper.add_n_argument(ap)
    args = ap.parse_args()

    injection.configure()
    settings_init = settings.use_base(config_values.functional)
    overrides = {
        'sleep_time': 0.1
    }
    if args.debug:
        overrides['log_level'] = logging.DEBUG
        overrides['log_to_console'] = True
    if args.fakes:
        overrides['use_fakes'] = True
    n_arg = arg_helper.get_overrides(args)
    if n_arg is not None and not args.fakes:
        overrides.update(n_arg)

    settings_init.add_overrides(overrides).configure()
    light_module.configure()
    machine.Machine().run(build_instructions())


if __name__ == '__main__':
    main()



import ast

from bardolph.vm.vm_codes import OpCode, Operator, Register

class Visitor:
    _class_to_op = {
        ast.UAdd: Operator.UADD,
        ast.USub: Operator.USUB,
        ast.Not: Operator.NOT,
        ast.And: Operator.AND,
        ast.Or: Operator.OR,
        ast.Add: Operator.ADD,
        ast.Mult: Operator.MUL,
        ast.Sub: Operator.SUB,
        ast.Div: Operator.DIV,
        ast.Eq: Operator.EQ,
        ast.NotEq: Operator.NOTEQ,
        ast.Lt: Operator.LT,
        ast.LtE: Operator.LTE,
        ast.Gt: Operator.GT,
        ast.GtE: Operator.GTE
    }

    def __init__(self, code_gen):
        self._code_gen = code_gen

    def expression(self, node):
        """ BinOp, UnaryOp, Num, Constant, or Name """
        if isinstance(node, ast.Expression):
            node = node.body
        fn_dict = {
            ast.BinOp: self._binop,
            ast.Compare: self._compare,
            ast.Constant: self._constant,
            ast.Name: self._name,
            ast.Num: self._num,
            ast.UnaryOp: self._unaryop
        }
        fn = fn_dict.get(type(node), self._nop)
        fn(node)

    def _binop(self, node):
        self.expression(node.left)
        self.expression(node.right)
        self._op(node.op)

    def _compare(self, node):
        self.expression(node.left)
        self.expression(node.comparators[0])
        self._code_gen.add_instruction(
            OpCode.OP,
            Visitor._class_to_op.get(type(node.ops[0]), None))

    def _constant(self, node):
        self._code_gen.add_instruction(OpCode.PUSH, node.n)

    def _name(self, node):
        name = node.id
        reg = Register.from_string(name)
        if reg is not None:
            self._code_gen.add_instruction(OpCode.PUSH, reg)
        else:
            self._code_gen.add_instruction(OpCode.PUSH, node.id)

    def _num(self, node):
        self._code_gen.add_instruction(OpCode.PUSHQ, node.n)

    def _unaryop(self, node):
        self.expression(node.operand)
        self._op(node.op)

    def _nop(self, node):
        print("parser error", node)

    def _op(self, op):
        self._code_gen.add_instruction(
            OpCode.OP, Visitor._class_to_op.get(type(op), None))


class ExprParser:
    def __init__(self, expr_string):
        try:
            self._tree = ast.parse(expr_string, mode="eval")
        except SyntaxError:
            self._tree = None

    def generate_code(self, code_gen) -> bool:
        """
        Generate the VM code to evaluate the expression. When that code is
        done, the result will be at the top of the stack.
        """
        if self._tree is None:
            return False

        visitor = Visitor(code_gen)
        visitor.expression(self._tree)
        return True



from enum import Enum

from bardolph.lib.auto_repl import auto
from bardolph.vm.vm_codes import Register


_RAW_RANGE = (0, 65535)


class UnitMode(Enum):
    LOGICAL = auto()
    RAW = auto()

def has_range(reg):
    return reg in (
        Register.BRIGHTNESS, Register.HUE, Register.KELVIN, Register.SATURATION)

def requires_conversion(reg):
    return reg in (
        Register.BRIGHTNESS, Register.DURATION, Register.HUE,
        Register.SATURATION, Register.TIME)

def get_range(reg):
    """
    Return the allowable range, in raw units, for a parameter.

    reg:
        token_type.TokenType designating the register to be set. May also
        be a string containing the name of the Enum.

    Returns:
        A tuple containing (minimum, maximum), or None if the parameter
        does not have a limited range of values.
    """
    reg = _string_check(reg)
    return (None, None) if reg in (
        Register.DURATION, Register.TIME) else _RAW_RANGE

def as_raw(reg, logical_value, use_float=False):
    """
    If necessary, converts to integer value that can be passed into the
    light API.

    Args:
        reg: TokenType corresponding to the register being set.
        logical_value: the number to be converted. May also
        be a string containing the name of the Enum.

    Returns:
        If no conversion is done, returns the incoming value untouched.
        Otherwise, an integer that corresponds to the logical value.
    """
    reg = _string_check(reg)
    if not requires_conversion(reg):
        return logical_value

    value = logical_value
    if reg == Register.HUE:
        if logical_value in (0.0, 360.0):
            value = 0.0
        else:
            value = (logical_value % 360.0) / 360.0 * 65535.0
    elif reg in (Register.BRIGHTNESS, Register.SATURATION):
        if logical_value == 100.0:
            value = 65535.0
        else:
            value = logical_value / 100.0 * 65535.0
    elif reg in (Register.DURATION, Register.TIME):
        value = logical_value * 1000.0

    if reg == Register.HUE and value > 65535.0:
        value %= 65536.0
    return value if use_float else round(value)

def as_logical(reg, raw_value):
    """If necessary, converts to floating-point logical value that
    typically apears in a script.

    Args:
        reg: TokenType corresponding to the register being set.
        raw_value: the number to be converted. May also
        be a string containing the name of the Enum.

    Returns:
        If no conversion is done, returns the incoming value untouched.
        Otherwise, a float that corresponds to the raw value.
    """
    reg = _string_check(reg)
    if not requires_conversion(reg):
        return raw_value

    value = raw_value
    if reg == Register.HUE:
        value = float(raw_value) / 65535.0 * 360.0
    elif reg in (Register.BRIGHTNESS, Register.SATURATION):
        if raw_value == 65535:
            value = 100.0
        else:
            value = float(raw_value) / 65535.0 * 100.0
    elif reg in (Register.DURATION, Register.TIME):
        value = raw_value / 1000.0
    return value

def _string_check(reg):
    return Register[reg.upper()] if isinstance(reg, str) else reg



from bardolph.vm.instruction import Instruction
from bardolph.vm.vm_codes import JumpCondition, OpCode, Operator, Register

class _JumpMarker:
    def __init__(self, inst, offset):
        self.jump = inst
        self.offset = offset
        self.has_else = False

class CodeGen:
    def __init__(self):
        self._code = []

    @property
    def program(self):
        return self._code

    @property
    def current_offset(self):
        return len(self._code)

    def clear(self):
        self._code.clear()

    def push(self, operand):
        self.add_instruction(CodeGen._push_op(operand), operand)

    def pop(self, operand):
        self.add_instruction(OpCode.POP, operand)

    def add_instruction(self, op_code, param0=None, param1=None):
        inst = Instruction(op_code, param0, param1)
        self._code.append(inst)
        return inst

    def add_list(self, inst_list):
        for code in inst_list:
            op_code = code[0]
            param0 = code[1] if len(code) > 1 else None
            param1 = code[2] if len(code) > 2 else None
            self.add_instruction(op_code, param0, param1)

    def addition(self, addend0, addend1) -> None:
        push0 = CodeGen._push_op(addend0)
        push1 = CodeGen._push_op(addend1)
        self.add_list([
            (push0, addend0),
            (push1, addend1),
            (OpCode.OP, Operator.ADD)
        ])

    def subtraction(self, minuend, subtrahend) -> None:
        # minuend - subtrahend
        # Leaves the difference on top of the stack.
        push = OpCode.PUSHQ if isinstance(
            subtrahend, (int, float)) else OpCode.PUSH
        self.add_list([
            (OpCode.PUSH, minuend),
            (push, subtrahend),
            (OpCode.OP, Operator.SUB)
        ])

    def decrement(self, dest) -> None:
        self.add_list([
            (OpCode.PUSH, dest),
            (OpCode.PUSHQ, 1),
            (OpCode.OP, Operator.SUB),
            (OpCode.POP, dest)
        ])

    def push_context(self, params):
        self.add_instruction(OpCode.JSR, params)

    def test_op(self, operator, op0, op1):
        """
        Generate code to perform a boolean operation on two operands. The
        generated code will set the RESULT register accordingly.

        If literal is True, then the second operand is a literal value and
        should be embedded verbatim in the instruction.
        """
        push = OpCode.PUSHQ if isinstance(op1, (int, float)) else OpCode.PUSH
        self.add_list([
            (OpCode.PUSH, op0),
            (push, op1),
            (OpCode.OP, operator),
            (OpCode.POP, Register.RESULT)
        ])

    def mark(self):
        return _JumpMarker(None, self.current_offset)

    def jump_back(self, marker):
        offset = marker.offset - self.current_offset
        self.add_instruction(OpCode.JUMP, JumpCondition.ALWAYS, offset)

    def if_start(self):
        inst = self.add_instruction(OpCode.JUMP, JumpCondition.IF_FALSE)
        return _JumpMarker(inst, self.current_offset)

    def if_else(self, marker):
        marker.has_else = True
        marker.jump.param1 = self.current_offset - marker.offset + 2
        inst = self.add_instruction(OpCode.JUMP, JumpCondition.ALWAYS)
        marker.jump = inst
        marker.offset = self.current_offset

    def if_end(self, marker):
        marker.jump.param1 = self.current_offset - marker.offset + 1

    def optimize(self):
        idem = (Register.NAME, Register.OPERAND, Register.TIME,
                Register.DURATION, Register.FIRST_ZONE, Register.LAST_ZONE)
        last_value = {}
        for inst in self._code:
            reg = inst.param1
            if inst.op_code == OpCode.MOVE:
                if reg in idem and reg in last_value:
                    del last_value[reg]
            elif inst.op_code == OpCode.MOVEQ and reg in idem:
                value = inst.param0
                if reg in last_value and value == last_value[reg]:
                    inst.nop()
                else:
                    last_value[reg] = value
            elif inst.op_code == OpCode.ROUTINE:
                last_value.clear()
        self._code = [inst for inst in self._code if inst.op_code != OpCode.NOP]

    @classmethod
    def _push_op(cls, oper):
        return OpCode.PUSHQ if isinstance(oper, (int, float)) else OpCode.PUSH



import logging
import time


from lifxlan.errors import WorkflowException


class Light:
    def __init__(self, lifx_light):
        self._impl = lifx_light
        self._name = lifx_light.get_label()
        self._group = lifx_light.get_group()
        self._location = lifx_light.get_location()
        self._multizone = lifx_light.supports_multizone()
        self._birth = time.time()
    
    def __repr__(self):
        fmt = 'Light(_name="{}", _group="{}", _location="{}", _multizone={}, '
        fmt += ' _birth={})'
        rep = fmt.format(
            self._name, self._group, self._location, self._multizone, 
            self._birth)
        return rep
            
    @property
    def group(self):
        return self._group
    
    @property
    def location(self):
        return self._location
    
    @property
    def name(self):
        return self._name
    
    @property
    def multizone(self):
        return self._multizone

    @property
    def age(self):
        #seconds
        return time.time() - self._birth
    
    def set_color(self, color, duration, rapid=True):
        try:
            self._impl.set_color(color, duration, rapid)
        except WorkflowException as ex:
            logging.warning("In set_color(): {}".format(ex))
    
    def get_color(self):
        try:
            return self._impl.get_color()
        except WorkflowException as ex:
            logging.warning("In get_color(): {}".format(ex))
        return [-1] * 4
    
    def set_zone_color(self, first_zone, last_zone, color, duration):
        try:
            self._impl.set_zone_color(first_zone, last_zone, color, duration)
        except WorkflowException as ex:
            logging.warning("In set_zone_color(): {}".format(ex))
        
    def get_color_zones(self, first_zone=None, last_zone=None):
        try:
            return self._impl.get_color_zones(first_zone, last_zone)
        except WorkflowException as ex:
            logging.warning("In get_color_zones(): {}".format(ex))
        
    def set_power(self, power, duration, rapid=True):
        try:
            return self._impl.set_power(power, duration, rapid)
        except WorkflowException as ex:
            logging.warning("In set_power(): {}".format(ex))
    
    def get_power(self):
        try:
            return self._impl.get_power()
        except WorkflowException as ex:
            logging.warning("In get_power(): {}".format(ex))
        return -1



class Series:
    def __init__(self, first=0, delta=0):
        self._first = first
        self._delta = delta
        self._iter = iter(SeriesIter(self).values())

    def init(self, first, delta):
        self.__init__(first, delta)
        self._iter = iter(SeriesIter(self).values())
        
    @property
    def first(self):
        return self._first
    
    @property
    def delta(self):
        return self._delta
   
    def next(self):
        return self._iter.__next__()


class SeriesIter:
    def __init__(self, series):
        self._first = series.first
        self._current = series.first
        self._delta = series.delta
        
    def values(self):
        self._current = self._first
        yield self._first

        while True:
            self._current += self._delta
            yield self._current



from enum import Enum

from bardolph.controller.units import UnitMode
from bardolph.lib.auto_repl import auto
from bardolph.vm.vm_codes import LoopVar, OpCode
from bardolph.vm.vm_codes import Operator
from bardolph.vm.vm_codes import Register

from .expr_parser import ExprParser
from .token_types import TokenTypes

class LoopType(Enum):
    INFINITE = auto()
    COUNTED = auto()
    CYCLE = auto()
    UNTIL = auto()
    WHILE = auto()


class LoopParser:
    def __init__(self, parser):
        self._loop_type = None
        self._parser = parser
        self._index_var = None
        self._counter = None

    def repeat(self, code_gen, call_context) -> bool:
        self._next_token()
        if not self._detect_loop_type(code_gen):
            return False
        if not self._pre_loop(code_gen, call_context):
            return False
        loop_top = code_gen.mark()
        if not self._loop_test(code_gen):
            return False
        if self._loop_type != LoopType.INFINITE:
            exit_loop_marker = code_gen.if_start()
        if not (self._loop_body() and self._loop_post(code_gen)):
            return False
        code_gen.jump_back(loop_top)
        if self._loop_type != LoopType.INFINITE:
            code_gen.if_end(exit_loop_marker)
        code_gen.add_instruction(OpCode.END_LOOP)
        return True

    def _detect_loop_type(self, code_gen) -> bool:
        code_gen.add_instruction(OpCode.LOOP)
        self._loop_type = LoopType.INFINITE
        if self._current_token_type == TokenTypes.WHILE:
            self._loop_type = LoopType.WHILE
            self._next_token()
        elif self._parser.at_rvalue():
            self._loop_type = LoopType.COUNTED
        return True

    def _pre_loop(self, code_gen, call_context) -> bool:
        if self._loop_type != LoopType.COUNTED:
            return True
        if not self._parser.rvalue(LoopVar.COUNTER):
            return False
        if self._current_token_type != TokenTypes.WITH:
            return True
        self._next_token()
        if not self._init_index_var(call_context):
            return False
        if self._current_token_type == TokenTypes.FROM:
            self._next_token()
            if not self._index_var_range(code_gen):
                return False
        elif self._current_token_type == TokenTypes.CYCLE:
            self._loop_type = LoopType.CYCLE
            self._next_token()
            if not self._cycle_var_range(code_gen):
                return False
        else:
            return self._parser.token_error(
                'Needed "from" or "cycle", got "{}"')
        return True

    def _init_index_var(self, call_context) -> bool:
        if self._current_token_type != TokenTypes.NAME:
            return self._parser.token_error('Needed variable name, got "{}"')
        self._index_var = self._current_token
        call_context.add_variable(self._index_var)
        return self._next_token()

    def _index_var_range(self, code_gen) -> bool:
        if not self._parser.rvalue(LoopVar.FIRST):
            return False
        if self._current_token_type != TokenTypes.TO:
            return self._parser.token_error('Needed "to", got "{}"')
        self._next_token()
        if not self._parser.rvalue(LoopVar.LAST):
            return False

        # increment = (last - first) / (count - 1)
        # If count == 1, skip to avoid division by 0.
        code_gen.add_instruction(OpCode.MOVE, LoopVar.FIRST, self._index_var)
        code_gen.test_op(Operator.NOTEQ, LoopVar.COUNTER, 1)
        marker = code_gen.if_start()
        code_gen.subtraction(LoopVar.LAST, LoopVar.FIRST)
        code_gen.subtraction(LoopVar.COUNTER, 1)
        code_gen.add_list([
            (OpCode.OP, Operator.DIV),
            (OpCode.POP, LoopVar.INCR)
        ])
        code_gen.if_end(marker)
        return True

    def _cycle_var_range(self, code_gen):
        if not self._parser.at_rvalue():
            code_gen.add_instruction(OpCode.MOVEQ, 0, LoopVar.FIRST)
        elif not self._parser.rvalue(LoopVar.FIRST):
            return False

        # increment = 360 / counter, or
        # increment = 65536 / counter
        code_gen.add_instruction(OpCode.MOVE, LoopVar.FIRST, self._index_var)
        code_gen.test_op(Operator.EQ, Register.UNIT_MODE, UnitMode.RAW)
        marker = code_gen.if_start()
        code_gen.push(65536)
        code_gen.if_else(marker)
        code_gen.push(360)
        code_gen.if_end(marker)
        code_gen.add_instruction(OpCode.PUSH, LoopVar.COUNTER)
        code_gen.add_instruction(OpCode.OP, Operator.DIV)
        code_gen.add_instruction(OpCode.POP, LoopVar.INCR)
        return True

    def _loop_test(self, code_gen) -> bool:
        if self._loop_type == LoopType.INFINITE:
            return True
        if self._loop_type == LoopType.WHILE:
            exp = ExprParser(self._current_token)
            if not exp.generate_code(code_gen):
                return False
            self._next_token()
        else:
            code_gen.add_list([
                (OpCode.PUSH, LoopVar.COUNTER),
                (OpCode.PUSHQ, 0),
                (OpCode.OP, Operator.GT)
            ])
        code_gen.add_instruction(OpCode.POP, Register.RESULT)
        return True

    def _loop_body(self) -> bool:
        return self._parser.command_seq()

    def _loop_post(self, code_gen) -> bool:
        if self._loop_type not in (LoopType.COUNTED, LoopType.CYCLE):
            return True
        code_gen.decrement(LoopVar.COUNTER)
        if self._index_var is not None:
            code_gen.addition(self._index_var, LoopVar.INCR)
            code_gen.add_instruction(OpCode.POP, self._index_var)
        return True

    @property
    def _current_token(self):
        return self._parser.current_token

    @property
    def _current_token_type(self):
        return self._parser.current_token_type

    def _next_token(self):
        return self._parser.next_token()



from collections import deque

from ..lib.symbol import Symbol, SymbolType
from ..lib.symbol_table import SymbolTable

class CallContext:
    _empty_table = SymbolTable()

    def __init__(self):
        self._stack = deque()
        self._in_routine = False
        self._globals = SymbolTable()
        self._stack.append(self._globals)

    def __contains__(self, name) -> bool:
        """
        Return True if the name exists as any type in the current context.
        """
        symbol_table = self.peek()
        if name in symbol_table:
            return True
        if symbol_table is not self._globals:
            return name in self._globals
        return False

    def clear(self) -> None:
        self._in_routine = False
        self._globals.clear()
        self._stack.clear()
        self._stack.append(self._globals)

    def enter_routine(self) -> None:
        self._in_routine = True

    def in_routine(self) -> bool:
        return self._in_routine

    def exit_routine(self) -> None:
        self._in_routine = False

    def push(self, symbol_table=None) -> None:
        if symbol_table is None:
            symbol_table = SymbolTable()
        self._stack.append(symbol_table)

    def pop(self) -> SymbolTable:
        assert len(self._stack) > 1
        self._stack.pop()
        return self.peek()

    def peek(self) -> SymbolTable:
        assert len(self._stack) > 0
        return self._stack[-1]

    def add_routine(self, routine) -> None:
        self._globals.add_symbol(routine.name, SymbolType.ROUTINE, routine)

    def add_variable(self, name, value=None) -> None:
        self.peek().add_symbol(name, SymbolType.VAR, value)

    def add_global(self, name, symbol_type, value) -> None:
        self._globals.add_symbol(name, symbol_type, value)

    def get_data(self, name) -> Symbol:
        return self.get_symbol_typed(name, (SymbolType.MACRO, SymbolType.VAR))

    def get_symbol(self, name) -> Symbol:
        """ Get a parameter from the top of the stack. If it's not there, check
        the globals. """
        symbol = self.peek().get_symbol(name)
        if symbol is not None:
            return symbol
        return self._globals.get_symbol(name)

    def get_symbol_typed(self, name, symbol_types) -> Symbol:
        symbol = self.get_symbol(name)
        if symbol is None or symbol.symbol_type not in symbol_types:
            return None
        return symbol

    def has_symbol(self, name) -> bool:
        return self.get_symbol(name) is not None

    def has_symbol_typed(self, name, * symbol_types) -> bool:
        symbol = self.get_symbol(name)
        return symbol is not None and symbol.symbol_type in symbol_types

    def get_routine(self, name):
        routine = None
        symbol = self._global_of_type(name, SymbolType.ROUTINE)
        if symbol is not None:
            routine = symbol.value
        return routine

    def get_macro(self, name):
        macro = self._global_of_type(name, SymbolType.MACRO)
        return None if macro is None else macro.value

    def _global_of_type(self, name, symbol_type):
        symbol = self._globals.get_symbol(name)
        if symbol is None or symbol.symbol_type != symbol_type:
            return None
        return symbol



#!/usr/bin/env python

import argparse
import logging

from bardolph.controller.routine import Routine
from bardolph.controller.units import UnitMode
from bardolph.lib.symbol_table import SymbolType
from bardolph.lib.time_pattern import TimePattern
from bardolph.vm.vm_codes import OpCode, Operand
from bardolph.vm.vm_codes import Register, SetOp

from . import lex
from .call_context import CallContext
from .code_gen import CodeGen
from .lex import Lex
from .loop_parser import LoopParser
from .expr_parser import ExprParser
from .token_types import TokenTypes


class Parser:
    def __init__(self):
        self._lexer = None
        self._error_output = ''
        self._call_context = CallContext()
        self._current_token_type = None
        self._current_token = None
        self._op_code = OpCode.NOP
        self._code_gen = CodeGen()
        self._command_map = {
            TokenTypes.ASSIGN: self._assignment,
            TokenTypes.BREAKPOINT: self._breakpoint,
            TokenTypes.DEFINE: self._definition,
            TokenTypes.GET: self._get,
            TokenTypes.IF: self._if,
            TokenTypes.NAME: self._call_routine,
            TokenTypes.OFF: self._power_off,
            TokenTypes.ON: self._power_on,
            TokenTypes.PAUSE: self._pause,
            TokenTypes.REGISTER: self._set_reg,
            TokenTypes.REPEAT: self._repeat,
            TokenTypes.SET: self._set,
            TokenTypes.UNITS: self._set_units,
            TokenTypes.WAIT: self._wait
        }
        self._token_trace = False

    def parse(self, input_string, optimize=False):
        self._call_context.clear()
        self._code_gen.clear()
        self._error_output = ''
        self._lexer = lex.Lex(input_string)
        self._next_token()
        success = self._script()
        self._lexer = None
        if success:
            if optimize:
                self._code_gen.optimize()
            return self._code_gen.program
        return None

    def load(self, file_name, optimize=False):
        logging.debug('File name: {}'.format(file_name))
        try:
            srce = open(file_name, 'r')
            input_string = srce.read()
            srce.close()
            return self.parse(input_string, optimize)
        except FileNotFoundError:
            logging.error('Error: file {} not found.'.format(file_name))
        except OSError:
            logging.error('Error accessing file {}'.format(file_name))

    def get_errors(self) -> str:
        return self._error_output

    @property
    def current_token(self) -> str:
        return self._current_token

    @property
    def current_token_type(self) -> TokenTypes:
        return self._current_token_type

    def next_token(self) -> bool:
        return self._next_token()

    def _script(self) -> bool:
        return self._body() and self._eof()

    def _body(self) -> bool:
        while self._current_token_type != TokenTypes.EOF:
            if not self._command():
                return False
        return True

    def _eof(self) -> bool:
        if self._current_token_type != TokenTypes.EOF:
            return self._trigger_error("Didn't get to end of file.")
        return True

    def _command(self) -> bool:
        return self._command_map.get(
            self._current_token_type, self._syntax_error)()

    def _set_reg(self):
        reg = Register.from_string(self._current_token)
        if reg is None:
            return self.token_error('Expected register, got "{}"')
        if reg == Register.TIME:
            return self._time()

        self._next_token()
        if self._current_token_type == TokenTypes.LITERAL_STRING:
            return self._string_to_reg(reg)
        return self.rvalue(reg)

    def _string_to_reg(self, reg) -> bool:
        if reg != Register.NAME:
            return self._trigger_error('Quoted value not allowed here.')
        self._add_instruction(OpCode.MOVEQ, self._current_token, reg)
        return self._next_token()

    def _set(self):
        return self._action(OpCode.COLOR)

    def _power_on(self):
        self._add_instruction(OpCode.MOVEQ, True, Register.POWER)
        return self._action(OpCode.POWER)

    def _power_off(self) -> bool:
        self._add_instruction(OpCode.MOVEQ, False, Register.POWER)
        return self._action(OpCode.POWER)

    def _action(self, op_code) -> bool:
        """ op_code: COLOR or POWER """
        self._op_code = op_code
        self._add_instruction(OpCode.WAIT)
        self._next_token()
        if self._current_token_type == TokenTypes.ALL:
            return self._all_operand()
        return self._operand_list()

    def _all_operand(self) -> bool:
        self._add_instruction(OpCode.MOVEQ, None, Register.NAME)
        self._add_instruction(OpCode.MOVEQ, Operand.ALL, Register.OPERAND)
        self._add_instruction(self._op_code)
        return self._next_token()

    def _operand_list(self) -> bool:
        """ For every operand in the list, issue the instruction in
        self._op_code. """
        if not self._operand():
            return False
        self._add_instruction(self._op_code)

        while self._current_token_type == TokenTypes.AND:
            self._next_token()
            if not self._operand():
                return False
            self._add_instruction(self._op_code)
        return True

    def _operand(self) -> bool:
        """
        Process a group, location, or light with an optional set of
        zones. Do this by populating the NAME and OPERAND registers.
        """
        if self._current_token_type == TokenTypes.GROUP:
            operand = Operand.GROUP
            self._next_token()
        elif self._current_token_type == TokenTypes.LOCATION:
            operand = Operand.LOCATION
            self._next_token()
        else:
            operand = Operand.LIGHT

        const_str = self._current_str()
        if const_str is not None:
            self._add_instruction(OpCode.MOVEQ, const_str, Register.NAME)
            self._next_token()
        elif self._current_token_type == TokenTypes.NAME:
            if not self._var_operand():
                return False
        else:
            return self.token_error('Needed a light, got "{}".')

        if self._current_token_type == TokenTypes.ZONE:
            if not self._zone_range():
                return False
            operand = Operand.MZ_LIGHT

        self._add_instruction(OpCode.MOVEQ, operand, Register.OPERAND)
        return True

    def _var_operand(self) -> bool:
        if not self._call_context.has_symbol_typed(
                self._current_token, SymbolType.VAR):
            return self.token_error('Not a variable: "{}"')
        self._add_instruction(OpCode.MOVE, self._current_token, Register.NAME)
        return self._next_token()

    def _zone_range(self) -> bool:
        if self._op_code != OpCode.COLOR:
            return self._trigger_error('Zones not supported for {}'.format(
                self._op_code.name.tolower()))
        self._next_token()
        return self._set_zones()

    def _set_zones(self, only_one=False):
        """
        Parses out one or two zone numbers. Generates instructions
        that populate the FIRST_ZONE and LAST_ZONE registers. If only one
        parameter is present, put None into LAST_ZONE.
        """
        if not self.at_rvalue():
            return self.token_error('Expected zone, got "{}"')
        if not self.rvalue(Register.FIRST_ZONE):
            return False
        if not only_one and self.at_rvalue():
            return self.rvalue(Register.LAST_ZONE)

        self._add_instruction(OpCode.MOVEQ, None, Register.LAST_ZONE)
        return True

    def _set_units(self) -> bool:
        self._next_token()
        mode = {
            TokenTypes.RAW: UnitMode.RAW,
            TokenTypes.LOGICAL: UnitMode.LOGICAL
        }.get(self._current_token_type, None)

        if mode is None:
            return self._trigger_error(
                'Invalid parameter "{}" for units.'.format(self._current_token))

        self._add_instruction(OpCode.MOVEQ, mode, Register.UNIT_MODE)
        return self._next_token()

    def _wait(self) -> bool:
        self._add_instruction(OpCode.WAIT)
        return self._next_token()

    def _get(self) -> bool:
        self._next_token()
        if not self.at_rvalue():
            return self.token_error('Needed light for get, got "{}".')
        if not self.rvalue(Register.NAME):
            return False

        if self._current_token_type == TokenTypes.ZONE:
            self._next_token()
            if not self._set_zones(True):
                return False
            operand = Operand.MZ_LIGHT
        else:
            operand = Operand.LIGHT

        self._add_instruction(OpCode.MOVEQ, operand, Register.OPERAND)
        self._add_instruction(OpCode.GET_COLOR)
        return True

    def _pause(self):
        self._add_instruction(OpCode.PAUSE)
        self._next_token()
        return True

    def _time(self):
        self._next_token()
        if self._current_token_type == TokenTypes.AT:
            self.next_token()
            return self._process_time_patterns()
        return self.rvalue(Register.TIME)

    def _process_time_patterns(self):
        time_pattern = self._current_time_pattern()
        if time_pattern is None:
            return self._time_spec_error()
        self._add_instruction(
            OpCode.TIME_PATTERN, SetOp.INIT, time_pattern)
        self._next_token()

        while self._current_token_type == TokenTypes.OR:
            self._next_token()
            time_pattern = self._current_time_pattern()
            if time_pattern is None:
                return self._time_spec_error()
            self._add_instruction(
                OpCode.TIME_PATTERN, SetOp.UNION, time_pattern)
            self._next_token()

        return True

    def _assignment(self) -> bool:
        self._next_token()
        if self._current_token_type != TokenTypes.NAME:
            return self.token_error('Expected name for assignment, got "{}"')
        dest_name = self._current_token
        self._next_token()
        if not self.rvalue(dest_name):
            return False
        self._call_context.add_variable(dest_name)
        return True

    def rvalue(self, dest=Register.RESULT) -> bool:
        """
        Consume the current token as an rvalue, generating the code to evaluate
        it and to move the result into dest.
        """
        move_inst = OpCode.MOVE
        value = self._current_constant()
        if value is not None:
            move_inst = OpCode.MOVEQ
        elif self._current_token_type == TokenTypes.NAME:
            name = self._current_token
            if self._call_context.has_symbol_typed(name, SymbolType.VAR):
                value = name
            else:
                return self.token_error('Unknown: "{}"')
        elif self._current_token_type == TokenTypes.EXPRESSION:
            parser = ExprParser(self._current_token)
            if not parser.generate_code(self._code_gen):
                return self.token_error('Error parsing expression "{}"')
            self._add_instruction(OpCode.POP, Register.RESULT)
            value = Register.RESULT
        elif self._current_token_type == TokenTypes.REGISTER:
            value = self._current_reg
        else:
            return self.token_error('Cannot use {} as a value.')

        self._add_instruction(move_inst, value, dest)
        return self._next_token()

    def at_rvalue(self) -> bool:
        if self._current_token_type not in (
                TokenTypes.EXPRESSION,
                TokenTypes.LITERAL_STRING,
                TokenTypes.NAME,
                TokenTypes.NUMBER):
            return False
        if self._current_token_type != TokenTypes.NAME:
            return True
        return not self._call_context.has_symbol_typed(
            self.current_token, SymbolType.ROUTINE)

    def _definition(self) -> bool:
        self._next_token()
        if self._current_token_type != TokenTypes.NAME:
            return self.token_error('Expected name for definition, got: {}')
        name = self._current_token

        self._next_token()
        if self._detect_routine_start():
            if self._call_context.get_routine(name) is not None:
                return self.token_error('Already defined: "{}"')
            return self._routine_definition(name)

        return self._macro_definition(name)

    def _detect_routine_start(self) -> bool:
        """
        If a definition is followed by "with", "begin", a keyword corresponding
        to a command, or the name of an existing routine, it's defining a new
        routine and not a variable.
        """
        if self._call_context.get_routine(self._current_token) is not None:
            return True
        return self._current_token_type in (
            TokenTypes.BEGIN, TokenTypes.WITH,
            TokenTypes.REGISTER, TokenTypes.SET)

    def _macro_definition(self, name):
        """
        Process a "define" where an alias for a value is being created. This
        symbol exists at compile time. This means a define cannot refer to a
        parameter in a routine. The symbol has global scope, even if it is
        defined inside a routine.
        """
        value = self._current_literal()
        if value is None:
            value = self._call_context.get_macro(self._current_token)
        if value is None:
            return self.token_error('Macro needs constant, got "{}"')
        self._call_context.add_global(name, SymbolType.MACRO, value)
        self._add_instruction(OpCode.CONSTANT, name, value)
        return self._next_token()

    def _routine_definition(self, name):
        if self._call_context.in_routine():
            return self._trigger_error('Nested definition not allowed.')

        self._call_context.enter_routine()
        self._call_context.push()
        self._add_instruction(OpCode.ROUTINE, name)

        routine = Routine(name)
        self._call_context.add_routine(routine)
        if self._current_token_type == TokenTypes.WITH:
            self._next_token()
            if not self._param_decl(routine):
                return False
        result = self.command_seq()
        self._call_context.pop()
        self._add_instruction(OpCode.END, name)

        self._call_context.exit_routine()
        return result

    def _param_decl(self, routine):
        """
        The parameter declarations for the routine are not included in the
        generated code. Declarations are used only at compile time.
        """
        name = self._current_token
        routine.add_param(name)
        self._call_context.add_variable(name)
        self._next_token()
        while self._current_token_type == TokenTypes.AND:
            self._next_token()
            name = self._current_token
            if routine.has_param(name):
                self.token_error('Duplicate parameter name: "{}"')
                return None
            routine.add_param(name)
            self._call_context.add_variable(name)
            self._next_token()
        return True

    def command_seq(self):
        if self._current_token_type != TokenTypes.BEGIN:
            return self._command()
        return self.compound_command()

    def compound_command(self):
        self._next_token()
        while self._current_token_type != TokenTypes.END:
            if self._current_token_type == TokenTypes.EOF:
                return self._trigger_error('End of file before "end".')
            if not self._command():
                return False
        return self._next_token()

    def _call_routine(self):
        routine = self._call_context.get_routine(self._current_token)
        if routine is None:
            return self.token_error('Unknown name: "{}"')

        self._next_token()
        for param_name in routine.params:
            if not self.rvalue():
                return False
            self._add_instruction(OpCode.PARAM, param_name, Register.RESULT)

        self._add_instruction(OpCode.JSR, routine.name)
        return True

    def _if(self) -> bool:
        self._next_token()
        if self._current_token_type != TokenTypes.EXPRESSION:
            return self.token_error('Unable to use as condition: "{}"')
        parser = ExprParser(self._current_token)
        if not parser.generate_code(self._code_gen):
            return self.token_error('Error parsing expression "{}"')
        self._add_instruction(OpCode.POP, Register.RESULT)
        marker = self._code_gen.if_start()
        self._next_token()
        if not self.command_seq():
            return False
        if self.current_token_type == TokenTypes.ELSE:
            self._code_gen.if_else(marker)
            self.next_token()
            if not self.command_seq():
                return False
        self._code_gen.if_end(marker)
        return True

    def _repeat(self):
        return LoopParser(self).repeat(self._code_gen, self._call_context)

    def _add_instruction(self, op_code, param0=None, param1=None):
        return self._code_gen.add_instruction(op_code, param0, param1)

    def _add_message(self, message):
        self._error_output += '{}\n'.format(message)

    def _trigger_error(self, message):
        full_message = 'Line {}: {}'.format(
            self._lexer.get_line_number(), message)
        self._add_message(full_message)
        return False

    def _current_literal(self):
        """
        Interpret the current token as a literal and return its value.
        If the current token doesn't contain a literal, return None.
        """
        value = None
        if self._current_token_type == TokenTypes.NUMBER:
            if Lex.INT_REGEX.match(self._current_token):
                value = int(self._current_token)
            else:
                value = float(self._current_token)
        elif self._current_token_type == TokenTypes.LITERAL_STRING:
            value = self._current_token
        elif self._current_token_type == TokenTypes.TIME_PATTERN:
            value = TimePattern.from_string(self._current_token)
            if value is None:
                self._time_spec_error()
        return value

    def _current_constant(self):
        """
        Interpret the current token as either a literal or macro, and return
        its value, which is known at compile time. If the token is an undefined
        name, return None.
        """
        value = self._current_literal()
        if value is not None:
            return value
        if self._current_token_type != TokenTypes.NAME:
            return None
        return self._call_context.get_macro(self._current_token)

    def _current_int(self):
        value = self._current_constant()
        if isinstance(value, int):
            return value
        if isinstance(value, float):
            return round(value)
        return None

    def _current_float(self):
        value = self._current_constant()
        if isinstance(value, float):
            return value
        if isinstance(value, int):
            return float(value)
        return None

    def _current_str(self):
        value = self._current_constant()
        return value if isinstance(value, str) else None

    def _current_time_pattern(self) -> TimePattern:
        """
        Returns the current token as a time pattern. If the current token
        isn't a time pattern or Symbol, returns None.

        Only literals or macros. Can't yet pass a time pattern as a
        parameter to a routine.
        """
        if self._current_token_type == TokenTypes.TIME_PATTERN:
            return TimePattern.from_string(self._current_token)
        if self._current_token_type == TokenTypes.NAME:
            return self._call_context.get_macro(self._current_token)
        return None

    def _current_reg(self):
        if self._current_token_type != TokenTypes.REGISTER:
            return None
        return Register.from_string(self._current_token)

    def _next_token(self):
        (self._current_token_type,
         self._current_token) = self._lexer.next_token()
        if self._token_trace:
            logging.info(
                'Next token: "{}" ({})'.format(
                    self._current_token, self._current_token_type))
        return True

    def _breakpoint(self):
        self._code_gen.add_instruction(OpCode.BREAKPOINT)

    def token_error(self, message_format):
        return self._trigger_error(message_format.format(self._current_token))

    def _unimplementd(self):
        return self.token_error('Unimplemented at token "{}"')

    def _syntax_error(self):
        return self.token_error('Unexpected input "{}"')

    def _time_spec_error(self):
        return self.token_error('Invalid time specification: "{}"')


def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('file', help='name of the script file')
    arg_parser.add_argument(
        '-o', '--optimize', help='enable optimization', action='store_true')
    args = arg_parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(filename)s(%(lineno)d) %(funcName)s(): %(message)s')
    parser = Parser()
    output_code = parser.load(args.file, args.optimize)
    if output_code:
        inst_num = 0
        for inst in output_code:
            print('{:5d} {}'.format(inst_num, inst))
            inst_num += 1
    else:
        print("Error parsing: {}".format(parser.get_errors()))


if __name__ == '__main__':
    main()



import re

from bardolph.lib.time_pattern import TimePattern
from .token_types import TokenTypes


class Lex:
    EXPR_REGEX = re.compile(r'^\{.*?\}$')
    TOKEN_REGEX = re.compile(r'#.*$|".*?"|\{.*?\}|\S+')
    NAME_REGEX = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')
    NUMBER_REGEX = re.compile(r'^\-?[0-9]*\.?[0-9]+$')
    INT_REGEX = re.compile(r'^\-?[0-9]*$')
    REG_REGEX = re.compile(
        '^hue$|^saturation$|^brightness$|^duration$|^time$|^kelvin$')

    def __init__(self, input_string):
        self._lines = iter(input_string.split('\n'))
        self._line_num = 0
        self._tokens = None
        self._next_line()

    def _next_line(self):
        current_line = next(self._lines, None)
        if current_line is None:
            self._tokens = None
        else:
            self._line_num += 1
            self._tokens = self.TOKEN_REGEX.finditer(current_line)

    @classmethod
    def _unabbreviate(cls, token):
        return {
            'h': 'hue', 's': 'saturation', 'b': 'brightness', 'k': 'kelvin'
        }.get(token, token)

    def get_line_number(self):
        return self._line_num

    def _token_type(self, token):
        token_type = TokenTypes.__members__.get(token.upper(), None)
        if token_type is not None:
            return token_type

        pairs = (
            (self.EXPR_REGEX, TokenTypes.EXPRESSION),
            (self.REG_REGEX, TokenTypes.REGISTER),
            (TimePattern.REGEX, TokenTypes.TIME_PATTERN),
            (self.NUMBER_REGEX, TokenTypes.NUMBER),
            (self.NAME_REGEX, TokenTypes.NAME)
        )
        for reg_expr, token_type in pairs:
            if reg_expr.match(token):
                return token_type

        return TokenTypes.UNKNOWN

    def next_token(self):
        token_type = None
        while token_type is None:
            match = None if self._tokens is None else next(self._tokens, None)
            while match is None:
                self._next_line()
                if self._tokens is None:
                    return (TokenTypes.EOF, '')
                match = next(self._tokens, None)

            token = Lex._unabbreviate(match.string[match.start():match.end()])
            if token[0] != '#':
                if token[0] == '"':
                    token = token[1:-1]
                    token_type = TokenTypes.LITERAL_STRING
                else:
                    token_type = self._token_type(token)
                    if token_type == TokenTypes.EXPRESSION:
                        token = token[1:-1]

        return (token_type, token)
